#include <stdlib.h>
#include "irensei_rule_description.h"
#include "irensei_ai_descripter.h"
#include "game.h"
#include <stdio.h>
#include <vector>
#include <algorithm>
#include <windows.h>


using namespace Irensei;

//																		//
//									AI									//

class AI {
public:
	AI( const IRENSEI_RULE_DESCRIPTION & rule, Game::EBlackOrWhite side );

	int Think( IRENSEI_BOARD_POSITION & result, const IRENSEI_BOARD_POSITION * lastMove, HANDLE hout );

	int Undo( IRENSEI_BOARD_POSITION & lastMove, HANDLE hout );

	int Debug( IRENSEI_DEBUG_BOARD & dest, int idx, HANDLE hout );
	static int CountDebug( void )		{ return 2; };

private:
	AI();		// ftHgRXgN^𖳌

	Game game;
	Game::EBlackOrWhite myside;

	IRENSEI_DEBUG_BOARD legalbuf;
	IRENSEI_DEBUG_BOARD weightTable;

	struct Position {
		Position()	{}
		Position( int a, int b ) : x(a), y(b)	{}
		Position(const Position&o) : x(o.x), y(o.y)	{}
		int x;
		int y;
	};
};

AI::AI( const IRENSEI_RULE_DESCRIPTION & rule, Game::EBlackOrWhite side )
		:
	myside(side)
{
	game.SetRule( rule );
}

int AI::Think( IRENSEI_BOARD_POSITION & result, const IRENSEI_BOARD_POSITION * const lastMove, HANDLE hout )
{
	// ̎ł
	if ( lastMove )
		game.NewMove( lastMove->x, lastMove->y );

	// @̗
	std::vector< Position > legalMoves;
	legalMoves.reserve( 19*19 );
	for ( int y = 0; 19 > y; y++ ) {
		for ( int x = 0; 19 > x; x++ ) {
			legalbuf.stones[y][x] = game.CheckLegalMove( game.GetBoard(), x, y, myside );
			if ( legalbuf.stones[y][x] <= Game::EMResultWin )		// @H
				legalMoves.push_back( Position(x,y) );
		}
	}

/*{
 char msg[512];
 sprintf( msg, "%d", legalMoves.size() );
 MessageBox( 0, msg, msg, MB_OK );
}*/
	IrenseiPrintf( hout, "%d\n", legalMoves.size() );

	// @肪Γ
	if ( legalMoves.size() == 0 )
		return 0;

	// _őI
	memset( weightTable.stones, -128, sizeof(weightTable.stones) );
	{
		// Ȃׂɋ߂ł悤ɂ
		std::vector< int > cofs;	// d݂̗ݐ
		int total = 0;
		cofs.push_back(total);
		for ( std::vector<Position>::const_iterator itor = legalMoves.begin(); itor != legalMoves.end(); itor++ ) {
			int cof = abs(itor->x - 9) + abs(itor->y - 9);
			cof = 20 - cof;		// 2..20 ŁA傫ȂB܂A[ 0 ɂ͂ȂȂ
			weightTable.stones[ itor->y ][ itor->x ] = cof;
			cof = cof * cof;	// Q悵āAW傫
			total += cof;
			cofs.push_back(total);
		}
		int r = total * rand() / RAND_MAX;
		std::vector< int >::const_iterator i = std::lower_bound( cofs.begin(), cofs.end(), r );
		*reinterpret_cast<Position*>(&result) = legalMoves[ i - cofs.begin() ];
	}

	// ̎ł
	game.NewMove( result.x, result.y );

	return 1;
}

int AI::Undo( IRENSEI_BOARD_POSITION & lastMove, HANDLE hout )
{
	//@@@@
	return 0;
}

int AI::Debug( IRENSEI_DEBUG_BOARD & dest, int idx, HANDLE /*hout*/ )
{
	switch ( idx ) {
		case 0:
			memcpy( &dest, &legalbuf, sizeof(dest) );
			break;
		case 1:
			memcpy( &dest, &weightTable, sizeof(dest) );
			break;
		default:
			return 0;
	}
	return 1;
}


//																		//
//						DLL C^[tF[X							//

extern "C" __declspec(dllexport) int IrenseiGetName( char dest[IRENSEI_AI_NAME_LEN] )	{ strcpy( dest, "random(bcc sample)" ); return 0; }
extern "C" __declspec(dllexport) int IrenseiGetFirstVersion( void )		{ return 100; }
extern "C" __declspec(dllexport) int IrenseiGetLastVersion( void )		{ return 100; }

extern "C" __declspec(dllexport) int IrenseiInitialize(const IRENSEI_RULE_DESCRIPTION * rule, int side, int hout)
{
	if ( ! CheckIrenseiRuleDescriptionValidity( rule ) )
		return 0;

	srand(time(0));

	IrenseiPrintf( reinterpret_cast<HANDLE>(hout), "random(bcc) AI initialized.\n" );
	return reinterpret_cast<int>( new AI( *rule, static_cast<Game::EBlackOrWhite>(side) ) );
}

extern "C" __declspec(dllexport) int IrenseiFinalize(int param, int hout)
{
	delete reinterpret_cast<AI*>( param );
	IrenseiPrintf( reinterpret_cast<HANDLE>(hout), "random(bcc) AI finalized.\n" );
	return 1;
}

extern "C" __declspec(dllexport) int IrenseiThink(int param, IRENSEI_BOARD_POSITION * result, const IRENSEI_BOARD_POSITION * lastMove, int hout)
{
	return reinterpret_cast<AI*>( param )->Think( *result, lastMove, reinterpret_cast<HANDLE>(hout) );
}

/* @@@ temporary enabled
extern "C" __declspec(dllexport) int IrenseiUndo( int param, IRENSEI_BOARD_POSITION * lastMove, int hout )
{
}
*/

extern "C" __declspec(dllexport) int IrenseiDebug( int param, IRENSEI_DEBUG_BOARD * dest, int idx, int hout )
{
	return reinterpret_cast<AI*>( param )->Debug( *dest, idx, reinterpret_cast<HANDLE>(hout) );
}

extern "C" __declspec(dllexport) int IrenseiDebugCount( void )
{
	return AI::CountDebug();
}

